import asyncio
import sqlite3

from itertools import repeat
from types import TracebackType
from typing import (
    Optional,
    Type,
    Generator,
    Any,
    Iterable,
    Coroutine
)

import async_timeout

from .core import Connection, DatabasePath
from .cursor import Cursor
from .exceptions import PoolError
from .transaction import IsolationLevel
from .factory import Record


class ConnectionProxy(Connection):

    __slots__ = '_in_use'

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        self._in_use: Optional[asyncio.Future] = None

    def _acquire(self) -> None:
        self._in_use = asyncio.get_running_loop().create_future()

    def _release(self) -> None:
        if not self._in_use.done():
            self._in_use.set_result(None)
        self._in_use = None

    async def _wait_until_released(self) -> None:
        if self._in_use is not None:
            await self._in_use


def connect(
        database: DatabasePath,
        *,
        timeout: float = 5.0,
        detect_types: int = 0,
        isolation_level: IsolationLevel = 'DEFERRED',
        check_same_thread: bool = False,
        factory: Type[Connection] = sqlite3.Connection,
        cached_statements: int = 128,
        uri: bool = False,
        prefetch: int = 64
) -> ConnectionProxy:
    """Create and return a connection to the sqlite database."""

    def _connector() -> sqlite3.Connection:
        if isinstance(database, str):
            loc = database
        elif isinstance(database, bytes):
            loc = database.decode('utf-8')
        else:
            loc = str(database)

        return sqlite3.connect(
            database=loc,
            timeout=timeout,
            detect_types=detect_types,
            isolation_level=None,
            check_same_thread=check_same_thread,
            factory=factory,
            cached_statements=cached_statements,
            uri=uri
        )

    return ConnectionProxy(_connector, isolation_level, prefetch)


class PoolAcquireContext:

    __slots__ = ('_pool', '_timeout', '_conn')

    def __init__(self, pool: "Pool", timeout: Optional[float]) -> None:
        self._pool = pool
        self._timeout = timeout
        self._conn = None

    def _acquired(self) -> None:
        if self._conn is not None:
            raise PoolError('A connection is already acquired.')

    def _acquire(self) -> Coroutine[Any, Any, ConnectionProxy]:
        return self._pool._acquire(timeout=self._timeout)

    def _release(self) -> None:
        conn = self._conn
        self._conn = None
        self._pool.release(conn)

    async def __aenter__(self) -> ConnectionProxy:
        self._acquired()
        self._conn = await self._acquire()
        return self._conn

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        self._release()

    def __await__(self) -> ConnectionProxy:
        self._acquired()
        self._conn = yield from self._acquire().__await__()
        return self._conn


class Pool:
    """A connection pool.

    Connection pool can be used to manage a set of connections to the database.
    Connections are first acquired from the pool, then used, and then released
    back to the pool. Once a connection is released, it's reset to close all
    open cursors and other resources *except* prepared statements.

    Pools are created by calling :func: `asyncsqlite3.create_pool`.
    """

    __slots__ = (
        '_database', '_min_size', '_max_size',
        '_close_timeout', '_connect_kwargs',
        '_initialized', '_initializing',
        '_all_connections', '_queue', '_closed'
    )

    def __init__(
            self,
            database: DatabasePath,
            min_size: int,
            max_size: int,
            close_timeout: Optional[float],
            **kwargs: Any
    ) -> None:

        if max_size <= 0:
            raise ValueError('max_size is expected to be greater than zero')

        if min_size < 0:
            raise ValueError(
                'min_size is expected to be greater or equal to zero')

        if min_size > max_size:
            raise ValueError('min_size is greater than max_size')

        self._database = database
        self._min_size = min_size
        self._max_size = max_size
        self._close_timeout = close_timeout
        self._connect_kwargs = kwargs
        self._initialized = False
        self._initializing = False
        self._all_connections = []
        self._queue = asyncio.LifoQueue(maxsize=self.get_max_size())
        self._closed = False

    def _create_new_connection(self) -> ConnectionProxy:
        conn = connect(self._database, **self._connect_kwargs)

        self._all_connections.append(conn)

        return conn

    async def _get_new_connection_if_current_closed(self, conn: ConnectionProxy) -> ConnectionProxy:
        if conn.is_closed():
            self._all_connections.remove(conn)
            conn = await self._create_new_connection()
        return conn

    async def _acquire(self, *, timeout: Optional[float]) -> ConnectionProxy:
        if self.is_closed():
            raise PoolError('Pool is closed.')

        if len(self._all_connections) < self.get_max_size():
            conn = await self._create_new_connection()
            conn._acquire()
            return conn

        try:
            if timeout is not None:
                async with async_timeout.timeout(timeout):
                    conn = await self._queue.get()
            else:
                conn = await self._queue.get()
        except asyncio.TimeoutError:
            raise PoolError('There are no free connections in the pool.') from None
        else:
            conn = await self._get_new_connection_if_current_closed(conn)
            conn._acquire()
            return conn

    def acquire(self, *, timeout: Optional[float] = None) -> PoolAcquireContext:
        """Acquire a database connection from the pool."""
        return PoolAcquireContext(self, timeout)

    def release(self, conn: ConnectionProxy) -> None:
        """Release a database connection back to the pool."""
        if self.is_closed():
            raise PoolError('Pool is closed.')
        if conn not in self._all_connections:
            raise PoolError('Connection not found in the pool.')
        if conn in self._queue._queue:
            raise PoolError('The connection is already in the pool.')
        if conn._in_use is None:
            raise PoolError('The connection is not currently used by the pool.')

        conn._release()

        self._queue.put_nowait(conn)

    async def close(self) -> None:
        """Attempt to gracefully close all connections in the pool."""
        if self.is_closed():
            raise PoolError('Pool is closed.')

        try:
            if self._close_timeout is not None:
                async with async_timeout.timeout(self._close_timeout):
                    await self._wait_all_connections()
            else:
                await self._wait_all_connections()
            await self.terminate()
        except asyncio.TimeoutError:
            await self.terminate()
            raise

    async def terminate(self) -> None:
        """Terminate all connections in the pool."""
        if self.is_closed():
            raise PoolError('Pool is closed.')

        try:
            await asyncio.gather(*[conn.close() for conn in self._all_connections])
        finally:
            self._closed = True
            self._all_connections.clear()

    async def execute(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None
    ) -> Cursor:
        """Pool performs this operation using one of its connections.
        Other than that, it behaves identically to Connection.execute().
        """
        async with self.acquire() as conn:
            return await conn.execute(sql, parameters, timeout=timeout)

    async def executemany(
            self,
            sql: str,
            parameters: Iterable[Iterable[Any]],
            *,
            timeout: Optional[float] = None
    ) -> Cursor:
        """Pool performs this operation using one of its connections.
        Other than that, it behaves identically to Connection.executemany().
        """
        async with self.acquire() as conn:
            return await conn.executemany(sql, parameters, timeout=timeout)

    async def executescript(
            self,
            sql_script: str,
            *,
            timeout: Optional[float] = None
    ) -> Cursor:
        """Pool performs this operation using one of its connections.
        Other than that, it behaves identically to Connection.executescript().
        """
        async with self.acquire() as conn:
            return await conn.executescript(sql_script, timeout=timeout)

    async def fetchone(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None,
            row_factory: Optional[Type] = Record
    ) -> Optional[Record]:
        """Pool performs this operation using one of its connections.
        Other than that, it behaves identically to Connection.fetchone().
        """
        async with self.acquire() as conn:
            return await conn.fetchone(sql, parameters, timeout=timeout, row_factory=row_factory)

    async def fetchmany(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            size: Optional[int] = None,
            timeout: Optional[float] = None,
            row_factory: Optional[Type] = Record
    ) -> Iterable[Record]:
        """Pool performs this operation using one of its connections.
        Other than that, it behaves identically to Connection.fetchmany().
        """
        async with self.acquire() as conn:
            return await conn.fetchmany(sql, parameters, size=size, timeout=timeout, row_factory=row_factory)

    async def fetchall(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None,
            row_factory: Optional[Type] = Record
    ) -> Iterable[Record]:
        """Pool performs this operation using one of its connections.
        Other than that, it behaves identically to Connection.fetchall().
        """
        async with self.acquire() as conn:
            return await conn.fetchall(sql, parameters, timeout=timeout, row_factory=row_factory)

    def get_max_size(self) -> int:
        """Return the maximum allowed number of connections in this pool."""
        return self._max_size

    def get_min_size(self) -> int:
        """Return the maximum allowed number of connections in this pool."""
        return self._min_size

    def get_size(self) -> int:
        """Return the current number of idle connections in this pool."""
        return self._queue.qsize()

    def is_closed(self) -> bool:
        return not self._all_connections and self._closed

    async def _wait_all_connections(self) -> None:
        await asyncio.gather(*[conn._wait_until_released() for conn in self._all_connections])

    async def _initialization(self) -> Optional["Pool"]:
        """Connect to the sqlite database and put connection in pool."""
        if self._initialized:
            return
        if self._initializing:
            raise PoolError('Pool initialization is already in progress.')
        if self.is_closed():
            raise PoolError('Pool is closed.')
        self._initializing = True
        try:
            for _ in repeat(None, self.get_min_size()):
                self._queue.put_nowait(self._create_new_connection())

            await asyncio.gather(*self._all_connections)

            return self
        finally:
            self._initializing = False
            self._initialized = True

    def _format(self) -> str:
        return 'size={} min_size={} max_size={} closed={}'.format(
            self.get_size(), self.get_min_size(), self.get_max_size(), self.is_closed())

    def __repr__(self) -> str:
        return f'<Pool at {id(self):#x} {self._format()}>'

    def __str__(self) -> str:
        return f'<Pool {self._format()}>'

    def __await__(self) -> Generator[Any, None, "Pool"]:
        return self._initialization().__await__()

    async def __aenter__(self) -> "Pool":
        return await self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        await self.close()


def create_pool(
        database: DatabasePath,
        *,
        min_size: int = 10,
        max_size: int = 10,
        close_timeout: Optional[float] = None,
        **kwargs: Any
) -> Pool:
    """Create and return a connection pool."""
    return Pool(database, min_size, max_size,
                close_timeout, **kwargs)



# MIT License
# Copyright (c) 2022 Vladyslav49

from sqlite3 import (
    register_adapter,
    register_converter,
    sqlite_version,
    sqlite_version_info,
    Row
)

from .core import connect, Connection, Cursor
from .pool import create_pool, Pool
from .factory import Record
from .transaction import Transaction
from .exceptions import (
    Warning,
    Error,
    DatabaseError,
    DataError,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    TransactionError,
    PoolError
)

try:
    import uvloop
except ImportError:
    pass
else:
    uvloop.install()

__author__ = 'Vladyslav49'
__license__ = 'MIT'

__all__ = [
    'register_adapter',
    'register_converter',
    'sqlite_version',
    'sqlite_version_info',
    'Row',
    'connect',
    'Connection',
    'Cursor',
    'create_pool',
    'Pool',
    'Record',
    'Transaction',
    'Warning',
    'Error',
    'DatabaseError',
    'DataError',
    'IntegrityError',
    'InterfaceError',
    'InternalError',
    'NotSupportedError',
    'OperationalError',
    'ProgrammingError',
    'TransactionError',
    'PoolError'
]



import itertools
import asyncio
import sqlite3
import sys

from logging import getLogger
from functools import partial
from pathlib import Path
from types import TracebackType
from typing import (
    Union,
    Optional,
    Iterable,
    Any,
    Callable,
    Type,
    AsyncIterator,
    Generator
)
from warnings import warn
from threading import Thread
from queue import SimpleQueue, Empty

import async_timeout

from .cursor import Cursor
from .context import contextmanager
from .factory import Record
from .transaction import Transaction, IsolationLevel
from .exceptions import (
    Warning,
    Error,
    DatabaseError,
    DataError,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError
)

DatabasePath = Union[bytes, str, Path]

LOGGER = getLogger(__name__)

_connection_name_counter = itertools.count(1).__next__


class Connection:

    __slots__ = (
        '_name', '_conn', '_connector',
        '_isolation_level', '_prefetch',
        '_queue', '_closed'
    )

    def __init__(
            self,
            connector: Callable[[], sqlite3.Connection],
            isolation_level: IsolationLevel,
            prefetch: int
    ) -> None:
        self._name = f'asyncsqlite3-connection-{_connection_name_counter()}'

        self._conn: Optional[sqlite3.Connection] = None
        self._connector = connector

        self._isolation_level = isolation_level
        self._prefetch = prefetch

        self._queue = SimpleQueue()
        self._closed = False

    def _handler(self) -> None:
        """Execute function calls on a separate thread."""
        while not self.is_closed():
            # Continues running until all queue items are processed,
            # even after connection is closed (so we can finalize all futures)
            try:
                future, function = self._queue.get(timeout=0.1)
            except Empty:
                pass
            else:
                try:
                    LOGGER.debug('executing: %s', function)

                    try:
                        result = function()
                    except sqlite3.IntegrityError as error:
                        raise IntegrityError(error) from None
                    except sqlite3.NotSupportedError as error:
                        raise NotSupportedError(error) from None
                    except sqlite3.DataError as error:
                        raise DataError(error) from None
                    except sqlite3.InterfaceError as error:
                        raise InterfaceError(error) from None
                    except sqlite3.InternalError as error:
                        raise InternalError(error) from None
                    except sqlite3.ProgrammingError as error:
                        raise ProgrammingError(error) from None
                    except sqlite3.OperationalError as error:
                        raise OperationalError(error) from None
                    except sqlite3.DatabaseError as error:
                        raise DatabaseError(error) from None
                    except sqlite3.Error as error:
                        raise Error(error) from None
                    except sqlite3.Warning as error:
                        raise Warning(error) from None

                    LOGGER.debug('operation %s completed', function)

                    future.get_loop().call_soon_threadsafe(future.set_result, result)
                except BaseException as error:
                    LOGGER.debug('returning exception: %s', error)

                    future.get_loop().call_soon_threadsafe(future.set_exception, error)

    async def _put(self, func, *args, timeout=None, **kwargs):
        """Queue a function with the given arguments for execution."""
        function = partial(func, *args, **kwargs)

        future = asyncio.get_running_loop().create_future()

        self._queue.put((future, function))

        if timeout is not None:
            async with async_timeout.timeout(timeout):
                result = await future
        else:
            result = await future

        return result

    def _set_cursor_row_factory(
            self,
            cursor: Cursor,
            row_factory: Optional[Type]
    ) -> None:
        """Set row_factory for the cursor object.
        If row_factory is not passed (default value is False), set the connection row_factory.
        """
        if row_factory is False:
            cursor.row_factory = self.row_factory
        else:
            cursor.row_factory = row_factory

    @contextmanager
    async def cursor(
            self,
            *,
            prefetch: Optional[int] = None,
            row_factory: Optional[Type] = False
    ) -> Cursor:
        """Create an asyncsqlite3 cursor wrapping a sqlite3 cursor object."""
        cursor = Cursor(self, await self._put(self._conn.cursor), prefetch)
        self._set_cursor_row_factory(cursor, row_factory)
        return cursor

    @contextmanager
    async def execute(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None
    ) -> Cursor:
        """Helper to create a cursor and execute the given query."""
        if parameters is None:
            parameters = []
        cursor = await self._put(self._conn.execute, sql, parameters, timeout=timeout)
        return Cursor(self, cursor)

    @contextmanager
    async def executemany(
            self,
            sql: str,
            parameters: Iterable[Iterable[Any]],
            *,
            timeout: Optional[float] = None
    ) -> Cursor:
        """Helper to create a cursor and execute the given multiquery."""
        cursor = await self._put(self._conn.executemany, sql, parameters, timeout=timeout)
        return Cursor(self, cursor)

    @contextmanager
    async def executescript(
            self,
            sql_script: str,
            *,
            timeout: Optional[float] = None
    ) -> Cursor:
        """Helper to create a cursor and execute a user script."""
        cursor = await self._put(self._conn.executescript, sql_script, timeout=timeout)
        return Cursor(self, cursor)

    async def fetchone(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None,
            row_factory: Optional[Type] = False
    ) -> Optional[Record]:
        """Shortcut version of asyncsqlite3.Cursor.fetchone."""
        async with self.execute(sql, parameters, timeout=timeout) as cursor:
            self._set_cursor_row_factory(cursor, row_factory)
            return await cursor.fetchone()

    async def fetchmany(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            size: Optional[int] = None,
            timeout: Optional[float] = None,
            row_factory: Optional[Type] = False
    ) -> Iterable[Record]:
        """Shortcut version of asyncsqlite3.Cursor.fetchmany."""
        async with self.execute(sql, parameters, timeout=timeout) as cursor:
            self._set_cursor_row_factory(cursor, row_factory)
            return await cursor.fetchmany(size)

    async def fetchall(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None,
            row_factory: Optional[Type] = False
    ) -> Iterable[Record]:
        """Shortcut version of asyncsqlite3.Cursor.fetchall."""
        async with self.execute(sql, parameters, timeout=timeout) as cursor:
            self._set_cursor_row_factory(cursor, row_factory)
            return await cursor.fetchall()

    async def commit(self) -> None:
        """Commit the current transaction."""
        await self._put(self._conn.commit)

    async def rollback(self) -> None:
        """Roll back the current transaction."""
        await self._put(self._conn.rollback)

    async def close(self) -> None:
        """Complete queued queries/cursors and close the connection."""
        if not self.is_closed():
            try:
                await self._put(self._conn.close)
            finally:
                self._closed = True
                self._conn = None

    async def enable_load_extension(self, value: bool) -> None:
        await self._put(self._conn.enable_load_extension, value)  # type: ignore

    async def load_extension(self, path: str) -> None:
        await self._put(self._conn.load_extension, path)  # type: ignore

    async def set_progress_handler(
            self, handler: Callable[[], Optional[int]], n: int
    ) -> None:
        await self._put(self._conn.set_progress_handler, handler, n)

    async def set_trace_callback(self, handler: Callable) -> None:
        await self._put(self._conn.set_trace_callback, handler)

    async def create_function(
            self, name: str, num_params: int, func: Callable, deterministic: bool = False
    ) -> None:
        """
        Create user-defined function that can be later used
        within SQL statements. Must be run within the same thread
        that query executions take place so instead of executing directly
        against the connection, we defer this to `run` function.
        In Python 3.8 and above, if *deterministic* is true, the created
        function is marked as deterministic, which allows SQLite to perform
        additional optimizations. This flag is supported by SQLite 3.8.3 or
        higher, ``NotSupportedError`` will be raised if used with older
        versions.
        """
        if sys.version_info >= (3, 8):
            await self._put(
                self._conn.create_function,
                name,
                num_params,
                func,
                deterministic=deterministic,
            )
        else:
            if deterministic:
                warn(
                    "Deterministic function support is only available on "
                    'Python 3.8+. Function "{}" will be registered as '
                    "non-deterministic as per SQLite defaults.".format(name)
                )

            await self._put(self._conn.create_function, name, num_params, func)

    async def iterdump(self) -> AsyncIterator[str]:
        """
        Return an async iterator to dump the database in SQL text format.
        Example::
            async for line in db.iterdump():
                ...
        """
        for line in await self._put(self._conn.iterdump):
            yield line

    async def backup(
            self,
            target: Union["Connection", sqlite3.Connection],
            *,
            pages: int = 0,
            progress: Optional[Callable[[int, int, int], None]] = None,
            name: str = "main",
            sleep: float = 0.250
    ) -> None:
        """
        Make a backup of the current database to the target database.

        Takes either a standard sqlite3 or asyncsqlite3 Connection object as the target.
        """
        if isinstance(target, Connection):
            target = target._conn

        await self._put(
            self._conn.backup,
            target,
            pages=pages,
            progress=progress,
            name=name,
            sleep=sleep,
        )

    def transaction(
            self,
            isolation_level: Optional[IsolationLevel] = None,
            *,
            timeout: Optional[float] = None
    ) -> Transaction:
        """Gets a transaction object."""
        if isolation_level is None:
            isolation_level = self._isolation_level
        return Transaction(self, isolation_level, timeout)

    def is_closed(self) -> bool:
        return self._closed and self._conn is None

    @property
    def prefetch(self) -> int:
        return self._prefetch

    @prefetch.setter
    def prefetch(self, value: int) -> None:
        self._prefetch = value

    @property
    def in_transaction(self) -> bool:
        return self._conn.in_transaction

    @property
    def isolation_level(self) -> IsolationLevel:
        return self._isolation_level

    @isolation_level.setter
    def isolation_level(self, value: IsolationLevel) -> None:
        self._isolation_level = value

    @property
    def row_factory(self) -> Optional[Type]:
        return self._conn.row_factory

    @row_factory.setter
    def row_factory(self, factory: Optional[Type]) -> None:
        self._conn.row_factory = factory

    @property
    def text_factory(self) -> Type:
        return self._conn.text_factory

    @text_factory.setter
    def text_factory(self, factory: Type) -> None:
        self._conn.text_factory = factory

    @property
    def total_changes(self) -> int:
        return self._conn.total_changes

    async def _initialization(self) -> "Connection":
        """Connect to the sqlite database."""
        if self._conn is None:
            try:
                Thread(
                    target=self._handler,
                    name=self._name,
                    daemon=True
                ).start()

                self._conn = await self._put(self._connector)
                self.row_factory = Record
            except BaseException:
                self._closed = True
                self._conn = None
                raise

        return self

    def __repr__(self) -> str:
        return f'<Connection at {id(self):#x} {self._format()}>'

    def __str__(self) -> str:
        return f'<Connection {self._format()}>'

    def _format(self) -> str:
        return f'name={self._name!r} closed={self.is_closed()}'

    def __await__(self) -> Generator[Any, None, "Connection"]:
        return self._initialization().__await__()

    async def __aenter__(self) -> "Connection":
        return await self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        await self.close()


def connect(
        database: DatabasePath,
        *,
        timeout: float = 5.0,
        detect_types: int = 0,
        isolation_level: IsolationLevel = 'DEFERRED',
        check_same_thread: bool = False,
        factory: Type[Connection] = sqlite3.Connection,
        cached_statements: int = 128,
        uri: bool = False,
        prefetch: int = 64
) -> Connection:
    """Create and return a connection to the sqlite database."""

    def _connector() -> sqlite3.Connection:
        if isinstance(database, str):
            loc = database
        elif isinstance(database, bytes):
            loc = database.decode('utf-8')
        else:
            loc = str(database)

        return sqlite3.connect(
            database=loc,
            timeout=timeout,
            detect_types=detect_types,
            isolation_level=None,
            check_same_thread=check_same_thread,
            factory=factory,
            cached_statements=cached_statements,
            uri=uri
        )

    return Connection(_connector, isolation_level, prefetch)



import sys

from enum import Enum
from types import TracebackType
from typing import Optional, Type, TYPE_CHECKING

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

import async_timeout

if TYPE_CHECKING:
    from .core import Connection

from .exceptions import TransactionError

IsolationLevel = Literal['DEFERRED', 'IMMEDIATE', 'EXCLUSIVE']


class TransactionState(Enum):
    NEW = 0
    STARTED = 1
    COMMITTED = 2
    ROLLEDBACK = 3
    FAILED = 4


class Transaction:
    """
    Asyncio Transaction for sqlite3.
    """

    __slots__ = (
        '_conn', '_isolation_level', '_managed',
        '_state', '_timeout_handler'
    )

    def __init__(
            self,
            conn: "Connection",
            isolation_level: IsolationLevel,
            timeout: Optional[float]
    ) -> None:

        if timeout is not None:
            timeout = async_timeout.timeout(timeout)

        self._conn = conn
        self._isolation_level = isolation_level
        self._timeout_handler = timeout
        self._managed = False
        self._state = TransactionState.NEW

    def _check_state(self, operation: str) -> None:
        if self._state is not TransactionState.STARTED:
            if self._state is TransactionState.NEW:
                raise TransactionError(
                    f'cannot {operation}; the transaction is not yet started')
            if self._state is TransactionState.COMMITTED:
                raise TransactionError(
                    f'cannot {operation}; the transaction is already committed')
            if self._state is TransactionState.ROLLEDBACK:
                raise TransactionError(
                    f'cannot {operation}; the transaction is already rolled back')
            if self._state is TransactionState.FAILED:
                raise TransactionError(
                    f'cannot {operation}; the transaction is in error state')

    async def start(self) -> None:
        """Enter the transaction or savepoint block."""
        if self._state is TransactionState.STARTED:
            raise TransactionError(
                'cannot start; the transaction is already started')

        try:
            async with self._conn.cursor() as cursor:
                await cursor.execute(f'BEGIN {self._isolation_level} TRANSACTION;')
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.STARTED

    async def commit(self) -> None:
        """Exit the transaction or savepoint block and commit changes."""
        self._check_state('commit')

        try:
            await self._conn.commit()
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.COMMITTED

    async def rollback(self) -> None:
        """Exit the transaction or savepoint block and rollback changes."""
        self._check_state('rollback')

        try:
            await self._conn.rollback()
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.ROLLEDBACK

    def _timeout_available(self) -> bool:
        return isinstance(self._timeout_handler, async_timeout.Timeout)

    @property
    def state(self) -> str:
        return self._state.name

    @property
    def isolation_level(self) -> IsolationLevel:
        return self._isolation_level

    async def __aenter__(self) -> "Transaction":
        if self._managed:
            raise TransactionError(
                'cannot enter context: already in an `async with` block')
        else:
            self._managed = True
        await self.start()

        if self._timeout_available():
            self._timeout_handler._do_enter()

        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        try:
            if exc_type is not None:
                await self.rollback()
            else:
                await self.commit()
        finally:
            self._managed = False
            if self._timeout_available():
                self._timeout_handler._do_exit(exc_type)

    def __repr__(self) -> str:
        return f'<Transaction at {id(self):#x} {self._format()}>'

    def __str__(self) -> str:
        return f'<Transaction {self._format()}>'

    def _format(self) -> str:
        return f'state={self._state.name!r} isolation_level={self._isolation_level!r}'



from sqlite3 import Row
from typing import Hashable, Any


class Record(Row):
    def get(self, name: Hashable, default: Any = None, /) -> Any:
        try:
            return self[name]
        except IndexError:
            return default

    def __repr__(self) -> str:
        result = '<Record'
        for key in self.keys():
            result += f' {key}={self[key]!r}'
        return f'{result}>'



from functools import wraps
from types import TracebackType
from typing import (
    Generator,
    Any,
    Optional,
    Type,
    Coroutine
)

from .cursor import Cursor


class ContextManager:

    __slots__ = ('_coro', '_obj')

    def __init__(self, coro: Coroutine[Any, Any, Any]):
        self._coro = coro

    def __await__(self) -> Generator[Any, None, "ContextManager"]:
        return self._coro.__await__()

    async def __aenter__(self) -> Cursor:
        self._obj = await self._coro
        return self._obj

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType]
    ) -> None:
        await self._obj.close()


def contextmanager(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        return ContextManager(method(self, *args, **kwargs))
    return wrapper



import sqlite3

from typing import (
    AsyncIterator,
    Any,
    Iterable,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Type
)

if TYPE_CHECKING:
    from .core import Connection

from .factory import Record


class Cursor:

    __slots__ = ('_prefetch', '_conn', '_cursor', '_closed', '_put')

    def __init__(
            self,
            conn: "Connection",
            cursor: sqlite3.Cursor,
            prefetch: Optional[int] = None
    ) -> None:
        if prefetch is None:
            self._prefetch = conn.prefetch
        else:
            self._prefetch = prefetch
        self._conn = conn
        self._cursor = cursor
        self._closed = False
        self._put = self._conn._put

    async def __aiter__(self) -> AsyncIterator[Record]:
        """Async iterator."""
        while True:
            rows = await self.fetchmany(self.prefetch)
            if not rows:
                return
            for row in rows:
                yield row

    async def execute(
            self,
            sql: str,
            parameters: Optional[Iterable[Any]] = None,
            *,
            timeout: Optional[float] = None
    ) -> "Cursor":
        """Execute the given query."""
        if parameters is None:
            parameters = []
        await self._put(self._cursor.execute, sql, parameters, timeout=timeout)
        return self

    async def executemany(
            self,
            sql: str,
            parameters: Iterable[Iterable[Any]],
            *,
            timeout: Optional[float] = None
    ) -> "Cursor":
        """Execute the given multiquery."""
        await self._put(self._cursor.executemany, sql, parameters, timeout=timeout)
        return self

    async def executescript(
            self,
            sql_script: str,
            *,
            timeout: Optional[float] = None
    ) -> "Cursor":
        """Execute a user script."""
        await self._put(self._cursor.executescript, sql_script, timeout=timeout)
        return self

    async def fetchone(
            self,
            *,
            timeout: Optional[float] = None
    ) -> Optional[Record]:
        """Fetch a single row."""
        return await self._put(self._cursor.fetchone, timeout=timeout)

    async def fetchmany(
            self,
            size: Optional[int] = None,
            *,
            timeout: Optional[float] = None
    ) -> Iterable[Record]:
        """Fetch up to `cursor.arraysize` number of rows."""
        if size is None:
            size = self.arraysize
        return await self._put(self._cursor.fetchmany, size, timeout=timeout)

    async def fetchall(
            self,
            *,
            timeout: Optional[float] = None
    ) -> Iterable[Record]:
        """Fetch all remaining rows."""
        return await self._put(self._cursor.fetchall, timeout=timeout)

    async def close(self) -> None:
        """Close the cursor."""
        await self._put(self._cursor.close)
        self._closed = True

    @property
    def prefetch(self) -> int:
        return self._prefetch

    @prefetch.setter
    def prefetch(self, value: int) -> None:
        self._prefetch = value

    @property
    def rowcount(self) -> int:
        return self._cursor.rowcount

    @property
    def lastrowid(self) -> int:
        return self._cursor.lastrowid

    @property
    def arraysize(self) -> int:
        return self._cursor.arraysize

    @arraysize.setter
    def arraysize(self, value: int) -> None:
        self._cursor.arraysize = value

    @property
    def description(self) -> Tuple[Tuple]:
        return self._cursor.description

    @property
    def connection(self) -> sqlite3.Connection:
        return self._cursor.connection

    @property
    def row_factory(self) -> Optional[Type]:
        return self._cursor.row_factory

    @row_factory.setter
    def row_factory(self, factory: Optional[Type]) -> None:
        self._cursor.row_factory = factory

    def _format(self) -> str:
        return f'connection={self._conn._name!r} closed={self._closed}'

    def __repr__(self) -> str:
        return f'<Cursor at {id(self):#x} {self._format()}>'

    def __str__(self) -> str:
        return f'<Cursor {self._format()}>'



from re import search, MULTILINE
from setuptools import setup, find_packages


with open('asyncsqlite3/_version.py') as f:
    version = search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), MULTILINE).group(1)


with open('README.md') as f:
    long_description = f.read()


setup(
    name='asyncsqlite3',
    version=version,
    url='https://github.com/Vladyslav49/asyncsqlite3',
    license='MIT',
    author='Vladyslav49',
    python_requires='>=3.7',
    description='Asynchronous driver for sqlite3.',
    long_description_content_type='text/markdown',
    long_description=long_description,
    install_requires=[
        'async_timeout>=4.0.1,<4.0.2',
        'typing_extensions>=4.3.0,<4.4.0; python_version < "3.8.0"',
    ],
    extras_require={
        'uvloop': [
            'uvloop>=0.16.0,<0.17.0',
        ],
    },
    include_package_data=False,
)
